#include <glib.h>
#include "df-object.h"


/*
 * General object stuff
 */

typedef struct _ObjectInfo ObjectInfo;
struct _ObjectInfo {
	guint				 type;
	char				*name;
	guint				 parent_type;
	DF__Object__init		 init_virtuals;
	DF__Object__init		 init_interfaces;
	DF__Object__init_data		 init_data;
	guint				 sizeof_object;
	guint				 sizeof_virtuals;
	guint				 sizeof_interfaces;
	gpointer			 virtuals;
	gpointer			 interfaces;
};

static GArray *types = NULL;

/* Private method prototypes */
static void DF__Object__destroy(DF__Object *self);
static void DF__Object__last_unref(DF__Object *self);

static void
init_type(DF__Object *object, guint type)
{
	ObjectInfo * info;

	g_assert(types != NULL && type < types->len);

	info = g_array_index(types, ObjectInfo *, type);

	g_assert(info != NULL);

	/* Go until we loop in which case this must be DF__Object */
	if(info->parent_type != type)
		init_type(object, info->parent_type);

	info->init_data(object);
}

DF__Object *
DF__Object__create(guint type)
{
	ObjectInfo * info;
	DF__Object * object;

	if( ! types ||
	    type >= types->len) {
		g_warning("Trying to create an unknown type %u", type);
		return NULL;
	}

	info = g_array_index(types,ObjectInfo *,type);

	g_assert(info != NULL);
	
	/* All objects are of the same size */
	object = g_malloc0(info->sizeof_object);
	object->_t.type = type;
	object->_v = info->virtuals;
	object->_i = info->interfaces;

	init_type(object, type);

	return object;
}

gboolean
DF__Object__is_type(guint type, guint istype)
{
	g_return_val_if_fail(types != NULL, FALSE);
	g_return_val_if_fail(type < types->len, FALSE);
	g_return_val_if_fail(istype < types->len, FALSE);

	if(type > istype) {
		/* Note: This can probably be done in a faster more efficent way */
		ObjectInfo *info = g_array_index(types, ObjectInfo *, type);
		do {
			/* If we hit DF__Object these types are ok */
			if(info->parent_type == info->type)
				return FALSE;
			info = g_array_index(types, ObjectInfo *, info->parent_type);
		} while(info->type != istype);
		return TRUE;
	} else if(type < istype) {
		return FALSE;
	} else 
		return TRUE;
}

gboolean
DF__Object__is_type_warn(guint type, guint istype)
{
	g_return_val_if_fail(types != NULL, FALSE);
	g_return_val_if_fail(type < types->len, FALSE);
	g_return_val_if_fail(istype < types->len, FALSE);

	if(DF__Object__is_type(type, istype)) {
		return TRUE;
	} else {
		ObjectInfo *info = g_array_index(types, ObjectInfo *, type);
		ObjectInfo *isinfo = g_array_index(types, ObjectInfo *, istype);
		g_warning("Invalid cast from '%s' to '%s'", info->name, isinfo->name);
		return FALSE;
	}
}

gpointer
DF__Object__get_virtuals(guint type)
{
	ObjectInfo * info;

	if( ! types ||
	    type >= types->len) {
		g_warning("Trying to access an unknown type %u", type);
		return NULL;
	}

	info = g_array_index(types, ObjectInfo *, type);

	g_assert(info != NULL);

	return info->virtuals;
}

gpointer
DF__Object__get_interfaces(guint type)
{
	ObjectInfo * info;

	if( ! types ||
	    type >= types->len) {
		g_warning("Trying to access an unknown type %u", type);
		return NULL;
	}

	info = g_array_index(types, ObjectInfo *, type);

	g_assert(info != NULL);

	return info->interfaces;
}

static void
init_class(ObjectInfo * iinfo, guint type)
{
	ObjectInfo * info;

	g_assert(types != NULL && type < types->len);

	info = g_array_index(types, ObjectInfo *, type);

	g_assert(info != NULL);

	/* Go until we loop in which case this must be DF__Object */
	if(info->parent_type != type)
		init_class(iinfo, info->parent_type);

	info->init_virtuals(iinfo->virtuals);
	info->init_interfaces(iinfo->interfaces);
}

guint
DF__Object__init_type(const char *name,
		      guint parent_type,
		      DF__Object__init init_virtuals,
		      DF__Object__init init_interfaces,
		      DF__Object__init_data init_data,
		      guint sizeof_object,
		      guint sizeof_virtuals,
		      guint sizeof_interfaces)
{
	guint type;
	ObjectInfo *info = g_new0(ObjectInfo, 1);

	if( ! types)
		types = g_array_new(FALSE, FALSE, sizeof(ObjectInfo *));

	type = types->len;

	info->type = type;
	info->name = g_strdup(name);
	info->parent_type = parent_type;
	info->init_virtuals = init_virtuals;
	info->init_interfaces = init_interfaces;
	info->init_data = init_data;
	info->sizeof_object = sizeof_object;
	info->sizeof_virtuals = sizeof_virtuals;
	info->sizeof_interfaces = sizeof_interfaces;
	info->virtuals = g_malloc0(sizeof_virtuals);
	info->interfaces = g_malloc0(sizeof_interfaces);

	g_array_append_val(types, info);

	init_class(info, type);

	return type;
}

/*
 * Object definition stuff
 */

/*
 * Constructor
 */
DF__Object *
DF__Object__construct(DF__Object *self)
{
	/* No extra construction neccessary */
	return self;
}

/*
 * Public methods
 */
void
DF__Object__ref(DF__Object *self)
{
	g_return_if_fail(self != NULL);
	g_return_if_fail(DF__IS_A(self, DF__Object));
	g_return_if_fail(self->ref_count > 0);

	self->ref_count ++;
}

void
DF__Object__unref(DF__Object *self)
{
	g_return_if_fail(self != NULL);
	g_return_if_fail(DF__IS_A(self, DF__Object));
	g_return_if_fail(self->ref_count > 0);

	if(self->ref_count > 1)
		self->ref_count --;
	else
		DF__Object__last_unref(self);
}

/*
 * Private methods
 */
static void
DF__Object__destroy(DF__Object *self)
{
	DF__CALL_VIRTUAL(self->_v, void, destroy, (self), 0);
}

static void
DF__Object__last_unref(DF__Object *self)
{
	DF__CALL_VIRTUAL(self->_v, void, last_unref, (self), 0);
}

/*
 * Virtual Methods
 */
static void
last_unref(DF__Object *self)
{
	g_return_if_fail(self->ref_count > 0);

	if(self->ref_count == 1)
		DF__Object__destroy(self);

	self->ref_count --;

	if(self->ref_count == 0)
		g_free(self);
}

/*
 * Init functions
 */
static void
init_virtuals(DF__Object__Virtuals *virtuals)
{
	virtuals->destroy = NULL;
	virtuals->last_unref = last_unref;
}

static void
init_interfaces(DF__Object__Interfaces *interfaces)
{
}

static void
init_data(DF__Object *self)
{
	self->ref_count = 1;
}

/*
 * Definitions macro
 */
DF__DEFINITIONS(DF__Object, DF__Object, init_virtuals, init_interfaces, init_data)



